#if !defined(_RANDOM_GEN_H_)
#define _RANDOM_GEN_H_


#ifdef WIN32
	#include <windows.h>
#else
	#include <time.h>
#endif

#include "cxmisc.h"

/*
   Multiply-with-carry generator is used here:
   temp = ( A*X(n) + carry )
   X(n+1) = temp mod (2^32)
   carry = temp / (2^32)
*/
#define  ICV_RNG_NEXT(x)    ((uint64)(unsigned)(x)*1554115554 + ((x) >> 32))
#define  ICV_CVT_FLT(x)     (((unsigned)(x) >> 9)|CV_1F)

class CvRandomGenerator
{
public:
	CvRandomGenerator()
	{
		m_rngs=0;m_rng_size=0;
	}
	~CvRandomGenerator()
	{
		if(m_rngs) cvFree(&m_rngs);
	}

	int init(int d)
	{
		if(m_rngs || d<=0)
			return 0;

		int i;
		const int cvrng_buf_size = cvAlign(sizeof(CvRNG)*d,CV_MALLOC_ALIGN);
		unsigned char * buffer = (unsigned char *)cvAlloc(cvrng_buf_size);
		
		if(buffer)
		{
			m_rngs=(CvRNG *)buffer;
			for(i=0;i<cvrng_buf_size;i++) buffer[i]=0;
			m_rng_size = d;
		}
		else
			return 0;

		for(i=0;i<d;i++ )
		{
			#ifdef WIN32
				int64 t = cvGetTickCount();
				m_rngs[i]= cvRNG(t+i);
				Sleep(t&0x1F);
			#else
				m_rngs[i]= cvRNG(time(0));
			#endif
		}
		return 1;
	}
	CvRNG * get_rng(){return m_rngs;}

	// random numbers of normal or Gaussian distribution
	int randn_float(float * r, int index, int size)
	{
		assert(m_rngs && size>=1 && index>=0 && index<m_rng_size);
		if(r)
		{
			icvRandn_0_1_32f_C1R(r,size,index);
			return 1;
		}
		else
			return 0;
	}
	float randn_float_one(int index)
	{
		assert(m_rngs && index>=0 && index<m_rng_size);
		float result;
		icvRandn_0_1_32f_C1R(&result,1,index);
		return result;
	}

protected:
	CvRNG * m_rngs;
	int m_rng_size;

private:
	// code is from OpenCV 1.0
	void icvRandn_0_1_32f_C1R( float* arr, int len, int index )
	{
		uint64 temp = m_rngs[index];
		int i;

		for( i = 0; i < len; i++ )
		{
			double x, y, v, ax, bx;

			for(;;)
			{
				x = ((int)temp)*1.167239e-9;
				temp = ICV_RNG_NEXT(temp);
				ax = fabs(x);
				v = 2.8658 - ax*(2.0213 - 0.3605*ax);
				y = ((unsigned)temp)*2.328306e-10;
				temp = ICV_RNG_NEXT(temp);
				if( y < v || ax < 1.17741 )
					break;
				bx = x;
				x = bx > 0 ? 0.8857913*(2.506628 - ax) : -0.8857913*(2.506628 - ax);
				if( y > v + 0.0506 )
					break;
				if( log(y) < .6931472 - .5*bx*bx )
				{
					x = bx;
					break;
				}
				if( log(1.8857913 - y) < .5718733-.5*x*x )
					break;

				do
				{
					v = ((int)temp)*4.656613e-10;
					x = -log(fabs(v))*.3989423;
					temp = ICV_RNG_NEXT(temp);
					y = -log(((unsigned)temp)*2.328306e-10);
					temp = ICV_RNG_NEXT(temp);
				}
				while( y+y < x*x );
				x = v > 0 ? 2.506628 + x : -2.506628 - x;
				break;
			}
			arr[i] = (float)x;
		}
		m_rngs[index] = temp;
	}
};
#endif